#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import json
import pickle
from io import StringIO

try:
    # yaml isn't standard with python.  It shouldn't be required if it
    # isn't used.
    import yaml
except ImportError:  # pragma: nocover
    yaml = None

from django.utils.encoding import smart_str
from django.utils.xmlutils import SimplerXMLGenerator

try:
    from django.core.serializers.json import (
        DateTimeAwareJSONEncoder as DjangoJSONEncoder,
    )
except ImportError:
    from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse
from django.core import serializers

from .utils import HttpStatusCode, Mimer


class Emitter:
    """
    Super emitter. All other emitters should subclass
    this one. It has the `construct` method which
    conveniently returns a serialized `dict`. This is
    usually the only method you want to use in your
    emitter. See below for examples.

    `RESERVED_FIELDS` was introduced when better resource
    method detection came, and we accidentially caught these
    as the methods on the handler. Issue58 says that's no good.
    """

    EMITTERS = {}
    RESERVED_FIELDS = {
        "read",
        "update",
        "create",
        "delete",
        "model",
        "anonymous",
        "allowed_methods",
        "fields",
        "exclude",
    }

    def __init__(self, payload, handler, anonymous=True):
        self.data = payload
        self.handler = handler
        self.anonymous = anonymous

        if isinstance(self.data, Exception):
            raise Exception from self.data

    def method_fields(self, handler, fields):
        if not handler:
            return {}

        ret = dict()

        for field in fields - Emitter.RESERVED_FIELDS:
            t = getattr(handler, str(field), None)

            if t and callable(t):
                ret[field] = t

        return ret

    def construct(self):
        if isinstance(self.data, HttpResponse):
            raise HttpStatusCode(self.data)
        else:
            return self.data

    def render(self, request):
        """
        This super emitter does not implement `render`,
        this is a job for the specific emitter below.
        """
        raise NotImplementedError("Please implement render.")

    def stream_render(self, request, stream=True):
        """
        Tells our patched middleware not to look
        at the contents, and returns a generator
        rather than the buffered string. Should be
        more memory friendly for large datasets.
        """
        yield self.render(request)

    @classmethod
    def get(cls, format):
        """
        Gets an emitter, returns the class and a content-type.
        """
        if format in cls.EMITTERS:
            return cls.EMITTERS.get(format)

        raise ValueError("No emitters found for type %s" % format)

    @classmethod
    def register(cls, name, klass, content_type="text/plain"):
        """
        Register an emitter.

        Parameters::
         - `name`: The name of the emitter ('json', 'xml', 'yaml', ...)
         - `klass`: The emitter class.
         - `content_type`: The content type to serve response as.
        """
        cls.EMITTERS[name] = (klass, content_type)

    @classmethod
    def unregister(cls, name):
        """
        Remove an emitter from the registry. Useful if you don't
        want to provide output in one of the built-in emitters.
        """
        return cls.EMITTERS.pop(name, None)


class XMLEmitter(Emitter):
    def _to_xml(self, xml, data):
        if isinstance(data, (list, tuple)):
            for item in data:
                xml.startElement("resource", {})
                self._to_xml(xml, item)
                xml.endElement("resource")
        elif isinstance(data, dict):
            for key, value in data.items():
                xml.startElement(key, {})
                self._to_xml(xml, value)
                xml.endElement(key)
        else:
            xml.characters(smart_str(data))

    def render(self, request):
        stream = StringIO()

        xml = SimplerXMLGenerator(stream, "utf-8")
        xml.startDocument()
        xml.startElement("response", {})

        self._to_xml(xml, self.construct())

        xml.endElement("response")
        xml.endDocument()

        return stream.getvalue()


Emitter.register("xml", XMLEmitter, "text/xml; charset=utf-8")
Mimer.register(lambda *a: None, ("text/xml",))


class JSONEmitter(Emitter):
    """
    JSON emitter, understands timestamps.
    """

    def decode(self, data):
        if isinstance(data, list):
            for num, val in enumerate(data):
                try:
                    data[num] = val.decode("utf8")
                except AttributeError:
                    pass
        return data

    def render(self, request):
        cb = request.GET.get("callback", None)
        assert cb is None, "JSONP Callbacks not suppoted"
        seria = json.dumps(
            self.decode(self.construct()),
            cls=DjangoJSONEncoder,
            ensure_ascii=False,
            indent=4,
        )

        # Callback
        # TODO: do we care about JSONP?
        # if cb and is_valid_jsonp_callback_value(cb):
        #     return '%s(%s)' % (cb, seria)

        return seria


Emitter.register("json", JSONEmitter, "application/json; charset=utf-8")
Mimer.register(json.loads, ("application/json",))


class YAMLEmitter(Emitter):
    """
    YAML emitter, uses `safe_dump` to omit the
    specific types when outputting to non-Python.
    """

    def render(self, request):
        return yaml.safe_dump(self.construct())


if yaml:  # Only register yaml if it was import successfully.
    Emitter.register("yaml", YAMLEmitter, "application/x-yaml; charset=utf-8")
    Mimer.register(lambda s: dict(yaml.safe_load(s)), ("application/x-yaml",))


class PickleEmitter(Emitter):
    """
    Emitter that returns Python pickled.
    """

    def render(self, request):
        return pickle.dumps(self.construct())


Emitter.register("pickle", PickleEmitter, "application/python-pickle")

"""
WARNING: Accepting arbitrary pickled data is a huge security concern.
The unpickler has been disabled by default now, and if you want to use
it, please be aware of what implications it will have.

Read more: http://nadiana.com/python-pickle-insecure

Uncomment the line below to enable it. You're doing so at your own risk.
"""


# Mimer.register(pickle.loads, ('application/python-pickle',))


class DjangoEmitter(Emitter):
    """
    Emitter for the Django serialized format.
    """

    def render(self, request, format="xml"):
        if isinstance(self.data, HttpResponse):
            return self.data
        elif isinstance(self.data, (int, str)):
            response = self.data
        else:
            response = serializers.serialize(format, self.data, indent=True)

        return response


Emitter.register("django", DjangoEmitter, "text/xml; charset=utf-8")



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import warnings
from pydoc import locate

import django
from decorator import decorator
from django import get_version as django_version
from django.http import HttpResponse


def format_error(error):
    return "Django Declarative APIs (Django %s) crash report:\n\n%s" % (
        django_version(),
        error,
    )


class rc_factory:
    """
    Status codes.
    """

    CODES = dict(
        ALL_OK=("OK", 200),
        CREATED=("Created", 201),
        ACCEPTED=("Accepted", 202),
        DELETED=("", 204),  # 204 says "Don't send a body!"
        BAD_REQUEST=("Bad Request", 400),
        UNAUTHORIZED=("Unauthorized", 401),
        FORBIDDEN=("Forbidden", 403),
        NOT_FOUND=("Not Found", 404),
        NOT_ACCEPTABLE=("Not Acceptable", 406),
        DUPLICATE_ENTRY=("Conflict/Duplicate", 409),
        NOT_HERE=("Gone", 410),
        INTERNAL_ERROR=("Internal Error", 500),
        NOT_IMPLEMENTED=("Not Implemented", 501),
        THROTTLED=("Throttled", 503),
    )

    def __getattr__(self, attr):
        """
        Returns a fresh `HttpResponse` when getting
        an "attribute". This is backwards compatible
        with 0.2, which is important.
        """
        try:
            (r, c) = self.CODES.get(attr)
        except TypeError:
            raise AttributeError(attr)

        if (r, c) == ("Forbidden", 401):  # pragma: nocover
            warnings.warn(
                "In future versions rc.FORBIDDEN will return 403 and rc.UNAUTHORIZED 401.",
                PendingDeprecationWarning,
            )
            warnings.warn(
                "Please change all your rc.FORBIDDEN for rc.UNAUTHORIZED",
                DeprecationWarning,
            )

        class HttpResponseWrapper(HttpResponse):
            """
            Wrap HttpResponse and make sure that the internal _is_string
            flag is updated when the _set_content method (via the content
            property) is called
            """

            def _set_content(self, content):
                """
                Set the _container and _is_string /
                _base_content_is_iter properties based on the type of
                the value parameter. This logic is in the construtor
                for HttpResponse, but doesn't get repeated when
                setting HttpResponse.content although this bug report
                (feature request) suggests that it should:
                http://code.djangoproject.com/ticket/9403
                """
                is_string = False
                if not isinstance(content, str) and hasattr(content, "__iter__"):
                    self._container = content
                else:
                    self._container = [content]
                    is_string = True
                if django.VERSION >= (1, 4):
                    self._base_content_is_iter = not is_string
                else:  # pragma: nocover
                    self._is_string = is_string

            try:
                content = property(HttpResponse._get_content, _set_content)
            except Exception:

                @HttpResponse.content.setter
                def content(self, content):
                    self._set_content(content)

        return HttpResponseWrapper(r, content_type="text/plain", status=c)


rc = rc_factory()


class FormValidationError(Exception):
    def __init__(self, form):
        self.form = form


class HttpStatusCode(Exception):
    def __init__(self, response):
        self.response = response


def coerce_put_post(request):
    """
    Django doesn't particularly understand REST.
    In case we send data over PUT, Django won't
    actually look at the data and load it. We need
    to twist its arm here.

    The try/except abominiation here is due to a bug
    in mod_python. This should fix it.
    """
    if request.method == "PUT":
        # Bug fix: if _load_post_and_files has already been called, for
        # example by middleware accessing request.POST, the below code to
        # pretend the request is a POST instead of a PUT will be too late
        # to make a difference. Also calling _load_post_and_files will result
        # in the following exception:
        #   AttributeError: You cannot set the upload handlers after the upload has been processed.
        # The fix is to check for the presence of the _post field which is set
        # the first time _load_post_and_files is called (both by wsgi.py and
        # modpython.py). If it's set, the request has to be 'reset' to redo
        # the query value parsing in POST mode.
        if hasattr(request, "_post"):
            del request._post
            del request._files

        try:
            request.method = "POST"
            request._load_post_and_files()
            request.method = "PUT"
        except AttributeError:  # pragma: nocover
            # TODO: Not quite sure how this block is supposed to be hit. Maybe something internal in
            # django.http.request.HttpRequest._load_post_and_files()? If request.method didn't exist then we'd
            # be running into issues with the above check
            request.META["REQUEST_METHOD"] = "POST"
            request._load_post_and_files()
            request.META["REQUEST_METHOD"] = "PUT"

        request.PUT = request.POST


class MimerDataException(Exception):
    """
    Raised if the content_type and data don't match
    """

    pass


class Mimer:
    TYPES = dict()

    def __init__(self, request):
        self.request = request

    def is_multipart(self):
        content_type = self.content_type()

        if content_type is not None:
            return content_type.lstrip().startswith("multipart")

        return False

    def loader_for_type(self, ctype):
        """
        Gets a function ref to deserialize content
        for a certain mimetype.
        """
        for loadee, mimes in Mimer.TYPES.items():
            for mime in mimes:
                if ctype.startswith(mime):
                    return loadee

    def content_type(self):
        """
        Returns the content type of the request in all cases where it is
        different than a submitted form - application/x-www-form-urlencoded
        """
        type_formencoded = "application/x-www-form-urlencoded"

        # Content-Type format:
        # http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.17
        ctype = self.request.META.get("CONTENT_TYPE", type_formencoded).split(";")[0]

        if type_formencoded in ctype:
            return None

        return ctype

    def translate(self):
        """
        Will look at the `Content-type` sent by the client, and maybe
        deserialize the contents into the format they sent. This will
        work for JSON, YAML, XML and Pickle. Since the data is not just
        key-value (and maybe just a list), the data will be placed on
        `request.data` instead, and the handler will have to read from
        there.

        It will also set `request.content_type` so the handler has an easy
        way to tell what's going on. `request.content_type` will always be
        None for form-encoded and/or multipart form data (what your browser sends.)
        """
        ctype = self.content_type()
        self.request.content_type = ctype

        if not self.is_multipart() and ctype:
            loadee = self.loader_for_type(ctype)

            if loadee:
                try:
                    data = self.request.body
                    # PY3: Loaders usually don't work with bytes:
                    data = data.decode("utf-8")
                    self.request.data = loadee(data)

                    # Reset both POST and PUT from request, as its
                    # misleading having their presence around.
                    self.request.POST = self.request.PUT = dict()
                except (TypeError, ValueError):
                    # This also catches if loadee is None.
                    raise MimerDataException
            else:
                self.request.data = None

        return self.request

    @classmethod
    def register(cls, loadee, types):
        cls.TYPES[loadee] = types

    @classmethod
    def unregister(cls, loadee):
        return cls.TYPES.pop(loadee)


def translate_mime(request):
    request = Mimer(request).translate()


def require_mime(*mimes):
    """
    Decorator requiring a certain mimetype. There's a nifty
    helper called `require_extended` below which requires everything
    we support except for post-data via form.
    """

    @decorator
    def wrap(f, self, request, *args, **kwargs):
        m = Mimer(request)
        realmimes = set()

        rewrite = {
            "json": "application/json",
            "yaml": "application/x-yaml",
            "xml": "text/xml",
            "pickle": "application/python-pickle",
        }

        for idx, mime in enumerate(mimes):
            realmimes.add(rewrite.get(mime, mime))

        if not m.content_type() in realmimes:
            return rc.BAD_REQUEST

        return f(self, request, *args, **kwargs)

    return wrap


require_extended = require_mime("json", "yaml", "xml", "pickle")


def locate_object(namespaced_name):
    name_components = namespaced_name.split(".")
    module_name = ".".join(name_components[:-1])
    module = locate(module_name)
    class_name = name_components[-1]
    return getattr(module, class_name)


def instantiate_class(namespaced_class_name, *args, **kwargs):
    cls = locate_object(namespaced_class_name)
    return cls(*args, **kwargs)


def preprocess_rsa_key(key_str):
    """Make Android and iOS RSA keys compatible with cryptography library.

    Android and iOS have slightly wonky, non-standard key formats. This updates
    the key to be standardized and compatible with pyca/cryptography.
    """
    if key_str.startswith("-----BEGIN CERTIFICATE"):
        key_str = key_str.replace("CERTIFICATE", "PUBLIC KEY")
    elif key_str.startswith("-----BEGIN RSA"):
        key_str = key_str.replace("BEGIN RSA", "BEGIN").replace("END RSA", "END")
    return key_str



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import json
from http import HTTPStatus

from django.test import TestCase, override_settings
from django.core.cache import cache

from . import testutils

from django_declarative_apis import models as auth_models


@override_settings(DECLARATIVE_ENDPOINT_TASKS_FORCE_SYNCHRONOUS=True)
class DeclarativeApisTestCase(TestCase):
    client_class = testutils.DeclarativeApisOAuthClient

    def setUp(self):
        self.consumer = auth_models.OauthConsumer.objects.create()

    def test_simplest_endpoint(self):
        self.client.get(
            "/simple", consumer=self.consumer, expected_status_code=HTTPStatus.OK
        )

    def test_dict_endpoint(self):
        resp = self.client.get(
            "/dict", consumer=self.consumer, expected_status_code=HTTPStatus.OK
        )
        data = resp.json()
        self.assertEqual(
            data,
            {
                "test": {
                    "pk": 1,
                    "int_field": 1,
                    "__expandable__": ["expandable_dict", "expandable_string"],
                },
                "deep_test": {
                    "test": {
                        "pk": 1,
                        "int_field": 1,
                        "__expandable__": ["expandable_dict", "expandable_string"],
                    }
                },
            },
        )

    def test_typed_parameter(self):
        response = self.client.get(
            "/simple?int_type_field=not_an_int",
            consumer=self.consumer,
            expected_status_code=HTTPStatus.BAD_REQUEST,
        )
        error = json.loads(response.content.decode("utf-8"))
        self.assertEqual(error["error_code"], 703)
        self.assertTrue("Invalid values for field(s): int_type_field")

    def test_skip_deferred_task(self):
        cache.set("deferred_task_called", False)
        self.client.get(
            "/simple?skip_task=True",
            consumer=self.consumer,
            expected_status_code=HTTPStatus.OK,
        )
        self.assertFalse(cache.get("deferred_task_called"))

    def test_run_deferred_task(self):
        cache.set("deferred_task_called", False)
        self.client.get(
            "/simple", consumer=self.consumer, expected_status_code=HTTPStatus.OK
        )
        self.assertTrue(cache.get("deferred_task_called"))

    def test_dict_field_endpoint(self):
        good_dict = {
            "length": 11,
            "description": "This is a description",
            "timestamp": "2022-10-24T00:00:00",
            "words": ["foo", "bar", "baz", "quux"],
        }
        test_data = [
            (good_dict, HTTPStatus.OK, "good dict"),
            ({}, HTTPStatus.OK, "empty_dict"),
            (list(good_dict), HTTPStatus.BAD_REQUEST, "list"),
            ("a string", HTTPStatus.BAD_REQUEST, "string"),
            (1337, HTTPStatus.BAD_REQUEST, "int"),
        ]

        for dct, expected_status, message in test_data:
            data = {"dict_type_field": dct}
            with self.subTest(message):
                response = self.client.post(
                    "/dictfield",
                    consumer=self.consumer,
                    data=data,
                    expected_status_code=expected_status,
                    content_type="application/json",
                )
                if expected_status == HTTPStatus.OK:
                    self.assertDictEqual(json.loads(response.content), data)

    def test_pydantic_field_endpoint(self):
        good_dict = {
            "length": 11,
            "description": "This is a description",
            "timestamp": "2022-10-24T00:00:00",
            "words": ["foo", "bar", "baz", "quux"],
        }
        test_data = [
            (good_dict, HTTPStatus.OK, "no errors"),
            ({**good_dict, "length": "eleven"}, HTTPStatus.BAD_REQUEST, "bad length"),
            (
                {**good_dict, "description": ["one", "two"]},
                HTTPStatus.BAD_REQUEST,
                "bad description",
            ),
            (
                {**good_dict, "timestamp": "2022-10-24T99:99:99"},
                HTTPStatus.BAD_REQUEST,
                "bad timestamp",
            ),
            (
                {**good_dict, "words": "foo bar baz quux"},
                HTTPStatus.BAD_REQUEST,
                "bad words",
            ),
        ]

        for dct, expected_status, message in test_data:
            data = {"pydantic_type_field": dct}
            with self.subTest(message):
                response = self.client.post(
                    "/pydanticfield",
                    consumer=self.consumer,
                    data=data,
                    expected_status_code=expected_status,
                    content_type="application/json",
                )
                if expected_status == HTTPStatus.OK:
                    self.assertDictEqual(json.loads(response.content), data)

    def test_nested_pydantic_field_endpoint(self):
        good_dict = {"b": "hello", "c": {"a": "world"}}
        test_data = [
            (good_dict, HTTPStatus.OK, "no errors"),
            ({**good_dict, "b": list("abc")}, HTTPStatus.BAD_REQUEST, "bad b"),
            ({**good_dict, "c": 11}, HTTPStatus.BAD_REQUEST, "bad c"),
            ({**good_dict, "c": {"a": list("abc")}}, HTTPStatus.BAD_REQUEST, "bad a"),
            ({**good_dict, "c": {}}, HTTPStatus.BAD_REQUEST, "missing a"),
            (
                {k: v for (k, v) in good_dict.items() if k != "b"},
                HTTPStatus.BAD_REQUEST,
                "missing b",
            ),
            (
                {k: v for (k, v) in good_dict.items() if k != "c"},
                HTTPStatus.BAD_REQUEST,
                "missing c",
            ),
        ]

        for dct, expected_status, message in test_data:
            data = {"nested_pydantic_type_field": dct}
            with self.subTest(message):
                response = self.client.post(
                    "/nestedpydanticfield",
                    consumer=self.consumer,
                    data=data,
                    expected_status_code=expected_status,
                    content_type="application/json",
                )
                if expected_status == HTTPStatus.OK:
                    self.assertDictEqual(json.loads(response.content), data)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-29 15:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [("contenttypes", "0002_remove_content_type_name")]

    operations = [
        migrations.CreateModel(
            name="OauthConsumer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.CharField(max_length=100, null=True)),
                ("key", models.CharField(db_index=True, max_length=18)),
                ("secret", models.CharField(max_length=32)),
                ("rsa_public_key_pem", models.TextField(blank=True, null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
            options={"abstract": False},
        )
    ]



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import json
import time
import urllib.parse
import logging

import oauth2

from oauthlib import oauth1 as oauthlib_oauth1
from oauthlib import common as oauthlib_common

from django.http import QueryDict

import django.test
from django.test import Client
from django.test.client import ClientHandler
from django.test.runner import DiscoverRunner

from django_declarative_apis import models

DEFAULT_CONTENT_TYPE = "application/x-www-form-urlencoded"
_ENCODERS = {
    DEFAULT_CONTENT_TYPE: lambda data: urllib.parse.urlencode(data),
    "application/json": lambda data: json.dumps(data),
}


class NoLoggingTestRunner(DiscoverRunner):

    """Don't log during tests."""

    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        logging.disable(logging.CRITICAL)
        return super().run_tests(test_labels, extra_tests, **kwargs)


class RequestCreatorMixin:
    def setUp(self):
        self.consumer = models.OauthConsumer.objects.create(name="test")
        self.consumer.save()

    def create_request(
        self,
        method="GET",
        path="/",
        url_fields=None,
        body=None,
        content_type=DEFAULT_CONTENT_TYPE,
        use_auth_header_signature=False,
    ):
        meth = getattr(django.test.RequestFactory(), method.lower())
        req = meth(
            path,
            _ENCODERS[content_type](body) if body else None,
            content_type=content_type,
        )
        req.consumer = self.consumer
        req.META["consumer"] = self.consumer

        if url_fields is not None:
            params = req.GET.copy()
            params.update(url_fields)
            req.GET = params

        # TODO: this should be better documented or on by default. If this is omitted or False, then requests will
        # fail in resource input validation (ResourceUpdateEndpointDefinition.validate_input)
        req.META["use_auth_header_signature"] = use_auth_header_signature
        OAuthClientHandler._build_request(req)

        return req


class OAuthClientHandler(ClientHandler):
    one_time_client_timestamp_override = None

    @classmethod
    def get_oauth_client_timestamp(cls):
        if cls.one_time_client_timestamp_override:
            result = cls.one_time_client_timestamp_override
            cls.one_time_client_timestamp_override = None
        else:
            result = int(time.time())
        return result

    @classmethod
    def _build_request(cls, request):
        consumer = request.META.get("consumer")

        # Make request params mutable so we can add authorization parameters.
        # We make the params immutable again before processing the request
        request.POST = request.POST.copy()
        request.GET = request.GET.copy()

        if consumer:
            if request.method == "POST":
                data = request.POST
            else:
                data = request.GET

            # This provides a way for us to override default values for testing.
            oauth_version = request.META.get("oauth_version", "1.0")
            oauth_nonce = request.META.get("oauth_nonce", oauth2.generate_nonce())
            oauth_client_timestamp = request.META.get(
                "oauth_timestamp", cls.get_oauth_client_timestamp()
            )

            rsa_key = request.META.get("rsa_key", None)
            oauth_signature_method = (
                oauthlib_oauth1.SIGNATURE_RSA
                if rsa_key
                else oauthlib_oauth1.SIGNATURE_HMAC
            )

            oauth_signature_data = {
                "oauth_version": oauth_version,
                "oauth_nonce": oauth_nonce,
                "oauth_timestamp": str(oauth_client_timestamp),
                "oauth_consumer_key": consumer.key,
                "oauth_signature_method": oauth_signature_method,
            }

            # collect ALL request parameters (original + OAuth) for signing
            all_request_parameters = data.copy()
            all_request_parameters.update(oauth_signature_data)

            if oauth_signature_method == oauthlib_oauth1.SIGNATURE_RSA:
                # use RSA-SHA1 signature method
                oauth1_client = oauthlib_oauth1.Client(
                    consumer.key,
                    signature_method=oauth_signature_method,
                    rsa_key=rsa_key,
                )

                oauth_request = oauthlib_common.Request(
                    request.build_absolute_uri(request.path),
                    http_method=request.method,
                    body=all_request_parameters,
                )

                oauth_signature_data[
                    "oauth_signature"
                ] = oauth1_client.get_oauth_signature(oauth_request)
            else:
                # use HMAC-SHA1 signature method
                oauth_signature_data.update({"oauth_signature_method": "HMAC-SHA1"})

                # Create oauth request object to compute signature
                oauth_request = oauth2.Request.from_consumer_and_token(
                    consumer,
                    None,
                    request.method,
                    request.build_absolute_uri(request.path),
                    all_request_parameters,
                    is_form_encoded=True,
                )

                # Add signature to django request
                signature_method = oauth2.SignatureMethod_HMAC_SHA1()
                oauth_request.sign_request(signature_method, consumer, None)
                oauth_signature_data["oauth_signature"] = oauth_request.get_parameter(
                    "oauth_signature"
                ).decode("utf-8")

            use_auth_header_signature = request.META.pop(
                "use_auth_header_signature", False
            )
            if use_auth_header_signature:
                auth_header_string = "OAuth " + ",".join(
                    [
                        '{0}="{1}"'.format(key, value)
                        for key, value in oauth_signature_data.items()
                    ]
                )
                request.META["HTTP_AUTHORIZATION"] = auth_header_string
            else:
                data.update(oauth_signature_data)

        # Recreate the GET and POST QueryDicts to make them immutable, as in production
        request.POST = QueryDict(request.POST.urlencode().encode("utf-8"))
        request.GET = QueryDict(request.GET.urlencode().encode("utf-8"))

    def get_response(self, request):
        self._build_request(request)
        return ClientHandler.get_response(self, request)


class DeclarativeApisOAuthClient(Client):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.handler = OAuthClientHandler()

    def request(self, expected_status_code=None, **kwargs):
        response = super().request(**kwargs)
        if expected_status_code is not None:
            if response.status_code != expected_status_code:
                print(response.content.decode("utf-8"))
                raise AssertionError(
                    "Status code, {}, did not match expected status code, {}".format(
                        response.status_code, expected_status_code
                    )
                )
        return response

    def put(self, path, data=None, expected_status_code=None, **kwargs):
        if "content_type" not in kwargs:
            data = data or {}
            kwargs["content_type"] = "application/x-www-form-urlencoded"
            data_qd = QueryDict(mutable=True)
            data_qd.update(data)
            data = data_qd.urlencode()

        return super().put(
            path, data, expected_status_code=expected_status_code, **kwargs
        )



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import json
import logging

from django.http import HttpResponse
from oauthlib.oauth1.rfc5849 import signature

from django_declarative_apis.authentication import (
    Authenticator,
    AuthenticatorHint,
    AuthenticationSuccess,
    AuthenticationFailure,
)
from django_declarative_apis.authentication.oauthlib.endpoint import (
    TweakedSignatureOnlyEndpoint,
)
from django_declarative_apis.authentication.oauthlib.oauth_errors import OAuthError
from django_declarative_apis.authentication.oauthlib.request_validator import (
    DjangoRequestValidator,
)

from . import oauth_errors

TwoLeggedOauth1Hint = AuthenticatorHint("OAuth ")


logger = logging.getLogger(__name__)


class TwoLeggedOauth1(Authenticator):
    def validate_missing_parameters(self, request, parameters=None):
        parameters = parameters or []

        """ Ensures that the request contains all required parameters. """
        params = [
            "oauth_consumer_key",
            "oauth_nonce",
            "oauth_signature",
            "oauth_signature_method",
            "oauth_timestamp",
        ]

        params.extend(parameters)

        try:
            collected_request_parameters = dict(
                signature.collect_parameters(
                    uri_query=request.GET.urlencode(),
                    body=request.POST.dict(),
                    headers=request.META,
                    exclude_oauth_signature=False,
                )
            )
            missing = list(
                param for param in params if param not in collected_request_parameters
            )
        except ValueError as e:
            error_message = str(e)
            logger.error(error_message)
            oauth_error = oauth_errors.build_error(error_message)
            request.auth_header = getattr(oauth_error, "auth_header", None)
            return oauth_error
        except Exception:  # pragma: nocover
            missing = params

        if missing:
            error_message = "parameter_absent:{}".format(",".join(missing))
            logger.error(error_message)
            missing_param_info = oauth_errors.build_error(error_message)
            request.auth_header = getattr(missing_param_info, "auth_header", None)
            return missing_param_info
        else:
            return True

    def is_authenticated(self, request):
        """Authenticates the requester using OAuth1.0a."""
        param_check = self.validate_missing_parameters(request)
        if isinstance(param_check, AuthenticationFailure):
            return param_check

        uri = request.build_absolute_uri(request.path)
        url_querystring = request.GET.urlencode()
        if url_querystring:
            uri += "?" + url_querystring

        body_form_data = request.POST.urlencode()

        headers = {k: v for (k, v) in request.META.items() if isinstance(v, str)}

        if body_form_data and "Content-Type" not in headers:  # pragma: nocover
            # TODO: is this only necessary because our test client sucks?
            # TODO (DB): is this still needed?
            headers["Content-Type"] = "application/x-www-form-urlencoded"

        # Verify request
        try:
            validator = DjangoRequestValidator(request)
            endpoint = TweakedSignatureOnlyEndpoint(validator)
            result, _ = endpoint.validate_request(
                uri, http_method=request.method, body=body_form_data, headers=headers
            )

            if result:
                request.consumer = validator.consumer

            error_message = (
                endpoint.validation_error_message or validator.validation_error_message
            )

            if result:
                return AuthenticationSuccess()
            else:
                return oauth_errors.build_error(error_message)
        except Exception as e:
            if hasattr(e, "message"):
                logger.error("Invalid oauthlib request: " + e.message)
            return AuthenticationFailure()

    def authenticate_header(self, request):
        """Returns the authentication header. If it does not exist, returns "Unknown OAuth Error"""
        return getattr(request, "auth_header", "Unknown OAuth Error")

    def challenge(self, oauth_error=None):
        """
        Returns a 401 response with a small bit on
        what OAuth is, and where to learn more about it.

        When this was written, browsers did not understand
        OAuth authentication on the browser side, and hence
        the helpful template we render. Maybe some day in the
        future, browsers will take care of this stuff for us
        and understand the 401 with the realm we give it.
        """

        if oauth_error is None:
            oauth_error = OAuthError()

        response = HttpResponse()
        response.status_code = 401

        response["WWW-Authenticate"] = oauth_error.auth_header

        content = {"error_code": 401, "error_message": oauth_error.detail}
        response.content = json.dumps(content)
        response["content-type"] = "application/json"

        return response



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import django.test
from django_declarative_apis import models


def test_consumer_getter(key):
    return key


uncallable = "foo"


class ModelsTestCase(django.test.TestCase):
    def test_get_consumer(self):
        with self.settings(
            DECLARATIVE_ENDPOINT_CONSUMER_GETTER="tests.test_models.test_consumer_getter"
        ):
            self.assertEqual(models.get_consumer("foo"), "foo")

    def test_get_consumer_non_callable(self):
        with self.settings(
            DECLARATIVE_ENDPOINT_CONSUMER_GETTER="tests.test_models.uncallable"
        ):
            self.assertRaises(TypeError, models.get_consumer, "foo")
